ver = "v1.16"
#v1.16 13-DEC-2014
#fixed recursion error when com port disc /error
#v1.15 12-DEC_2014
#added safecast upload via http/api

#v1.14 02-APR-2014
#fixing threads

#v1.13 30-OCT-2013
	#made email, logging, SQL all thread-based
	#increased SQL connect timeout to 10s
	#fixed Email to only send on CPMAvg intervals

#v1.12 29-SEP-2013
	#fixed email to only send on datatimer intervals

#v1.11 08-Sep-2013 NJC
#v1.1  06-Sep-2013 NJC
	#threading
#v1.02 06-Aug-2013 NJC
#v1.01 17-Jun-2013 NJC
#v1.0  25-Mar-2013 NJC  njc@njctech.com 



import time
import serial
import sys
import struct
import msvcrt
import traceback
import argparse
import logging
from time import gmtime, strftime
import pyodbc
import os
from colorama import Fore, Back, Style
from colorama import init
init()
import smtplib
import string
import ConfigParser
import csv
import Queue
import threading
import httplib, urllib, urllib2
import json


descstr = "Serial monitor application for GMC-300 geiger counter " + ver	#description 

#commandline arg parser
parser = argparse.ArgumentParser(description=descstr, epilog="'s','S' displays settings. 'CTRL+Q' quits.")
parser.add_argument("-v", "--verbose", action="store_true", dest="verbose", help="Additional messages pirinted to console")
parser.add_argument("-s", "--sql", action="store_true", dest="SQL_on", default=False, help="Enable SQL server data logging")
parser.add_argument("-e", "--email", action="store_true", dest="EMAIL_on", default=False, help="Enable E-mail notifications")
parser.add_argument("-l", "--datalog", action="store_true", dest="LOGon", default=False, help="Enable CSV data logging")
parser.add_argument("-p", "--post", action="store_true", dest="POST_on", default=False, help="Enable POST Uploading")
parser.add_argument("-c", dest="comport", type=int, required=True, help="select COM port to use")
args = parser.parse_args()

#verbose settings
if args.verbose:
	consolelevel	= "logging.DEBUG"
	filelevel		= "logging.DEBUG" 
else:
	consolelevel	= "logging.WARNING"
	filelevel		= "logging.INFO"

#setup logging for both file and console
logging.basicConfig(filename="geiger.log",level=eval(filelevel), format='%(asctime)s %(levelname)-8s %(message)s')
console = logging.StreamHandler(sys.stdout)
console.setLevel(eval(consolelevel))
consoleformat = logging.Formatter('\n%(message)s')
console.setFormatter(consoleformat)
logging.getLogger('').addHandler(console)

#read and process config file
Config = ConfigParser.ConfigParser()
Config.read("geiger.ini")
Config.sections()

def ConfigSectionMap(section):
    dict1 = {}
    options = Config.options(section)
    for option in options:
        try:
            dict1[option] = Config.get(section, option)
            if dict1[option] == -1:
				logging.debug("skip: %s" % option)
        except:
            logging.debug("exception on %s!" % option)
            dict1[option] = None
    return dict1

#data log settings
if args.LOGon:
	fieldnames = ['UTC','Count','CPM','uSv/hr','mR/hr']
	datalogFile = ConfigSectionMap("Logging")['logfile']
	d = open(datalogFile, "a")
	#datalog = csv.DictWriter(open(datalogFile,'w'), delimiter=',', lineterminator='\n', fieldnames=fieldnames)
	datalog = csv.writer(d)	
	datalogTimer = ConfigSectionMap("Logging")['logtimer']
	datalog_on = 1
	datalog.writerow(fieldnames,)
	datalog.writerow([strftime("%d-%b-%Y %H:%M:%S UTC", gmtime())])
else:
	datalog_on = 0

#program settings
heartbeat_ON	=	ConfigSectionMap("Settings")['heartbeaton']
heartbeat_OFF	=	ConfigSectionMap("Settings")['heartbeatoff']
CPMAvgS			=	int(ConfigSectionMap("Settings")['cpmavg'])

#tube data
CPMuSvhr		=	int(ConfigSectionMap("Settings")['cpmusvhr']) 		    # CPM = 1uSv/hr
CPMmRhr			=	int(ConfigSectionMap("Settings")['cpmmrhr'])			# CPM = 1mR/hr
bytemask		=	int(ConfigSectionMap("Settings")['bytemask'],16)		#mask to eliminate top two MSB per GQ datasheet

#SQL settings
if args.SQL_on:
	SQLsrv		=	ConfigSectionMap("SQL")['sqlserver']
	SQLdb		=	ConfigSectionMap("SQL")['sqldb']
	SQLtable	=	ConfigSectionMap("SQL")['sqltable']
	SQLuid		=	ConfigSectionMap("SQL")['sqlusername']
	SQLpwd		=	ConfigSectionMap("SQL")['sqlpassword']
	SQLtimer	=	ConfigSectionMap("SQL")['sqltimer']
	SQLcxnstr	=	"DRIVER={SQL Server};SERVER="+SQLsrv+";DATABASE="+SQLdb+";UID="+SQLuid+";PWD="+SQLpwd
	SQLexecstr	=	"insert into "+SQLtable+"(CPM,uSvhr,mRhr) values (?,?,?)"
	SQL_on		=	1	#turn SQL logging on
else:
	SQL_on		=	0	#turn SQL logging off

#POST Settings
if args.POST_on:
	POSTtimer	=	ConfigSectionMap("POST")['posttimer']
	POSTserver	= 	ConfigSectionMap("POST")['postserver']
	POSTquery	= 	ConfigSectionMap("POST")['postquery']
	POSTparams	=	ConfigSectionMap("POST")['postparams']
	
	POSTcpmparam=	ConfigSectionMap("POST")['postcpmparam']
	POSTdtparam	=	ConfigSectionMap("POST")['postdtparam']
	POSTdtformat=	ConfigSectionMap("POST")['postdtformat']
	POSTokcode	=	ConfigSectionMap("POST")['postokcode']
	POSTheaders = 	{"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
	POST_on		= 	1 	#turn POST on
else:
	POST_on		=	0 	#turn POST off


#email settings
if args.EMAIL_on:
	emailTO		=	ConfigSectionMap("Email")['emailto']
	emailFROM	=	ConfigSectionMap("Email")['emailfrom']
	emailSERVER =	ConfigSectionMap("Email")['emailserver']
	emailThresh =	int(ConfigSectionMap("Email")['emailthresh'])
	emailTimer	=	int(ConfigSectionMap("Email")['emailtimer'])
	emailLink	=	ConfigSectionMap("Email")['emaillink']
	
	EMAIL_on		=	1	#turn email alerting on
else:
	EMAIL_on		=	0	#turn email alerting off

#set comport
comport = (args.comport - 1)	#comport-1 (0 = COM1, 1 = COM2, etc)

#hard settings
exWaitS		= 5		#time to wait before retrying if exception occurs (s)
hbtimeout	= 5		#heartbeat timeout is detection threshold (s)

#vars
CPS			= []		#CPS array
CPM			= 0			#CPM
uSvhr		= 0			#uSv/hr
mRhr		= 0			#mR/hr
ser			= 0			#serial instance
count		= long(0)	#count counter


global quitFlag
quitFlag = False
global SQLThread
SQLThread = threading.Thread()
global POSTThread
POSTThread = threading.Thread()
global loggingThread
loggingThread = threading.Thread()
global EmailThread
EmailThread = threading.Thread()

#startup time stamp
now = time.gmtime(time.time())
startedUTC = time.strftime("%Y-%m-%d %H:%M:%S UTC", now)



def setup():
	#logging.debug("*************************************")
	#logging.debug("Startup")
	os.system('cls')	#clear screen (windows only)
	global ser			#serial object is global
	
	
	#setup COM port
	ser = serial.Serial(
			#port=comport, # 0 is COM1, use 1 for COM2 etc
			baudrate=57600,
			bytesize=serial.EIGHTBITS,
			parity=serial.PARITY_NONE,
			stopbits=serial.STOPBITS_ONE,
			xonxoff=0,
			rtscts=0,
			dsrdtr=0,
			timeout=1,
			writeTimeout=0
			)	
	displaySettings()

def displaySettings():	
	#msg = "Settings:\n"
	#msg += "1uSv/hr=" + str(CPMuSvhr) + "CPM " + "1mR/hr=" + str(CPMmRhr) + "CPM "
	print("\n" + descstr)	#print welcome msg 
	startedmsg = "Startup: " + startedUTC
	msg = "CPMavg:" + str(CPMAvgS) + "s"
	if datalog_on:
		msg += " Log:On/" + str(datalogTimer) + "s"
	else:
		msg += " Log:Off"
	if SQL_on:
		msg += " SQL:On/" + str(SQLtimer) + "s"
	else:
		msg += " SQL:Off"
	if EMAIL_on:
		msg += " Email:On/" + str(emailThresh) + "CPM/" + str(emailTimer/60) + "min"
	else:
		msg += " Email:Off"
	if POST_on:
		msg += " POST:On/" + str(POSTtimer) + "s"
	else:
		msg += " POST:Off"

	#log and display the settings
	logging.warning(startedmsg)
	logging.warning(msg+"\n")	
	#print("\nSettings:\n" + msg)
	#logging.INFO(CPMuSvhr)


def start():	
	while not quitFlag:					#check for quit
		if openport():					#try to open the port
			try:
				run()					#port is open and we can enter run loop
			except Exception,e:
				logging.error(str(e))
				
		try:		
			closeport()					#we finished run - so close the port
			if not quitFlag:			#wait a bit before retrying, unless we're quitting
				time.sleep(exWaitS)								
		except Exception,e:
			logging.error(str(e))		

		#now start the while loop over 	
		
def openport():
	try:
		# Open the serial port 
		ser.port = comport
		try:
			ser.open()
			if ser.isOpen():
				logging.info(ser.portstr+" opened.")
				logging.info("Enabling heartbeat signal.")
				ser.write(heartbeat_ON)		
				return True
			else:
				return False
		except IOError, e:	
			logging.error(str(e))		
			return False
	except Exception,e:
		logging.error(str(e))
		return False


def closeport():
	try:
		if ser.isOpen():
			logging.info("Disabling heartbeat signal.")
			ser.write(heartbeat_OFF)
			try: 
				ser.close()
			except IOError, e:
				logging.error(str(e))			
			if ser.isOpen():
				logging.error(ser.portstr+" could not be closed.")
			else:
				logging.info(ser.portstr+" closed.")
	except Exception,e:
		logging.error(str(e))	


def sendEmail(subj,message):
	SUBJECT = subj
	TO = emailTO
	FROM = emailFROM
	text = message
	BODY = string.join((
	        "From: %s" % FROM,
	        "To: %s" % TO,
	        "Subject: %s" % SUBJECT ,
	        "",
	        text
	        ), "\r\n")
	server = smtplib.SMTP(emailSERVER)
	logging.warning("Sending E-mail...")
	try:
		server.sendmail(FROM, [TO], BODY)
		server.quit()
	except Exception, e:
		logging.error("Error sending E-Mail")
		logging.debug(str(e))

def POST(CPM,timestamp):	
	paramstr =   '{' + POSTparams + ',' + '"' + POSTcpmparam + '"' + ':' + '"' + str(int(round(CPM))) + '"' + ',' + '"' + POSTdtparam + '"' + ':' + '"' + strftime(POSTdtformat, timestamp) + '"' +'}'	
	params = urllib.urlencode(json.loads(paramstr))	
	try:
		logging.debug("Sending POST data")	
		conn = httplib.HTTPConnection(POSTserver)
		#conn.set_debuglevel(1)
		conn.request("POST", POSTquery, params ,POSTheaders)
		response = conn.getresponse()
		#logging.debug(response.status)
		if(str(response.status) == POSTokcode):
			logging.debug("POST to " + POSTserver + POSTquery + " (" + str(response.status) + ":" + response.reason + ")")
		else:
			logging.error("POST failed " + POSTserver + POSTquery + " (" + str(response.status) + ":" + response.reason + ")")
	except Exception, e:
		logging.error("Error POSTing data to server")
		logging.debug(str(e))


def SQL_connect(CPM, uSvhr, mRhr):
	try:																					
		logging.debug("Writing to SQL DB")		
		#setup and write to SQL db
		cnxn = pyodbc.connect(SQLcxnstr)
		cnxn.timeout = 10
		cursor = cnxn.cursor()
		cursor.execute(SQLexecstr,CPM, uSvhr, mRhr)
		cnxn.commit()
		
	except Exception, e:									#catch error writing to SQL
		logging.error("Error connecting to SQL DB")
		logging.debug(str(e))

def log_write(logstr):
	try:																
		logging.debug("Writing to CSV log file")
		#write data to CSV
		datalog.writerow(logstr)	
		#d.flush()											#flush so it writes the data NOW
	except csv.Error as e:
		logging.error("Error writing to CSV log file")			#catch error writing to CSV
		logging.debug(str(e))

def quit():
	logging.error("Exiting...")
	logging.debug("Stopping Threads...")	
	global quitFlag				#access the global quitFlag
	quitFlag = True				#signal threads to stop
	mainThread.join()			#wait for main thread to finish
	if(SQLThread.is_alive()):
		SQLThread.join()
	if(POSTThread.is_alive()):
		POSTThread.join()
	if(EmailThread.is_alive()):
		EmailThread.join()
	if(loggingThread.is_alive()):
		loggingThread.join()
	closeport()					#close the COM port gracefully
	logging.debug("Exit!")		
	#sys.exit()					#exit
	#exit occurs when checkKey finishes
	
def checkKey():	
	while not quitFlag:
		if msvcrt.kbhit() == 1:					#check if key has been pressed (windows only)		
			charint = ord(msvcrt.getch())		#get the key		
			if charint == 17:					# 'CTRL+Q'
				quit()							
			elif charint == 115 or charint == 83: # 's' or 'S'
				displaySettings()	
		time.sleep(0.1)		

def run():		#this is the main function	
	
	t0 = time.clock()			#when we started run()
	t1 = t0						#stamp of last read
	elapsed	= -1				#elapsed seconds - this counts seconds since we got into run(), resets if an exception occurs	
	global count				#total count since program started, does NOT reset if exception occurs
	timeout = False				#set timeout flag to false
	timer = 0					#reset timeout timer
	emailSent = -999999			#set initial email timer to a large negative number to handle an alert cond before thresh s has elapsed
	isGood = 0					#reset isGood flag
	msgColor = "Back.RED"		#default color
	logging.debug("isGood:" + str(isGood))
	CPS[:]=[]					#clear the CPS array
	printokmesg = 1				#first time through indicate heartbeat is alive
	
	while not quitFlag:	
		waitBytes = 0	
		try:
			waitBytes = ser.inWaiting()
		except IOError, e:	
			logging.error(str(e))
			#break
			

		if waitBytes == 2:						#check if there are TWO bytes waiting			
			if printokmesg:								#if first time, indicate heartbeat is ok
				logging.info("Receiving data...")		#heartbeat is ok
				sys.stdout.write("\n")
				printokmesg = 0							#no further need to indicate heartbeat ok
			r = struct.unpack('!1H', ser.read(2))[0]	#read the two bytes and unpack binary as hex and reverse them
			t1 = time.clock()							#reset timer to current time		
			c = r & bytemask							#AND with bytemask in order to get rid of top two MSB (per GQ)
			CPS.append(c)								#add latest reading to the array
			if len(CPS) >= CPMAvgS:						#if array is full (based on CPMAvgS setting)				
				CPS.pop(0)								#remove oldest reading		
				if isGood == 0:							#if isGood flag is still 0 then					
					isGood = 1							#set isGood flag to 1
					msgColor = "Back.GREEN"				#change bg color to green
					logging.debug("isGood:" + str(isGood))								
			count += c									#add to global CPM counter			
			CPM = float(sum(CPS))/float(len(CPS))*60	#calculate CPM			
			uSvhr = round((float(CPM)/float(CPMuSvhr)),2)#calculate uSv/hr
			mRhr = round((float(CPM)/float(CPMmRhr)),3)	#calculate mR/hr

			#time calculations
			elapsed = int(time.clock() - t0)			# seconds elapsed since we entered run()				
			min = elapsed/60							#minutes
			hr  = min/60								#hours
			day = hr/24									#days
			secd = "%02d" % (elapsed % 60)				#seconds, fixed and ready for display
			mind = "%02d" % (min % 60)					#minutes, fixed and ready for display
			hrd = "%02d" % (hr % 24)					#hours, fixed and ready for display
			dayd = day									#days, fixed and ready for display
			
			sys.stdout.flush()							#flush display buffer, this allows the second-by-second updates to be written on the same line
			sys.stdout.write('\r')
			now = time.gmtime(time.time())
			UTC = time.strftime("%Y%m%d %H:%M:%S UTC", now)
			
			#prepare current reading info for display
			msg =  eval(msgColor) + "Count:" + str(count) + " " + str(int(round(CPM))) + "CPM" + " " + str(uSvhr) + "uSv/hr" + " " + str(mRhr) + "mR/hr" + " [" +str(dayd) + ":" + str(hrd) + ":" + str(mind) + ":" + str(secd) + "] " + UTC + Style.RESET_ALL
			
			#CSV data log
			if datalog_on and ((elapsed % (int(datalogTimer))) == 0 and isGood):
				#logstr = "'" + strftime("%d-%b-%Y %H:%M:%S UTC", gmtime()) + "','" + str(count) + "','" + str(int(round(CPM))) +"','" + str(uSvhr) + "','" + str(mRhr) + "'"
				logstr = [strftime("%d-%b-%Y %H:%M:%S UTC", gmtime()),str(count),str(int(round(CPM))),str(uSvhr),str(mRhr)]
				loggingThread = threading.Thread(target=log_write, args=(logstr,))
				loggingThread.daemon = False
				loggingThread.start()
				msg += "-"				#indicate CSV operation by displaying a '-'
				
			#SQL
			if SQL_on and ((elapsed % (int(SQLtimer))) == 0 and isGood):
				SQLThread = threading.Thread(target=SQL_connect, args=(CPM, uSvhr, mRhr))
				SQLThread.daemon = False 	
				SQLThread.start()
				msg += "+"				#indicate SQL operation by displaying a '+'
			#POST
			if POST_on and ((elapsed % (int(POSTtimer))) == 0 and isGood):
				POSTThread = threading.Thread(target=POST, args=(CPM,gmtime()))
				POSTThread.daemon = False
				POSTThread.start()
				msg += "p"

			#Email alert			
			if EMAIL_on and CPM > emailThresh and ((time.clock() - emailSent) > emailTimer) and ((elapsed % (int(CPMAvgS))) == 0 and isGood):
				logging.warning("CPM over threshhold:" + str(int(round(CPM))) + "/" + str(emailThresh))
				emailSubj = "Radiation Alert: "+ str(int(round(CPM))) + "CPM"
				emailMsg = strftime("%d-%b-%Y %H:%M:%S UTC", gmtime()) + "\n" + str(int(round(CPM)))+"CPM\n" + emailLink
				EmailThread = threading.Thread(target=sendEmail, args=(emailSubj,emailMsg))
				EmailThread.daemon = False 	
				EmailThread.start()
				emailSent = time.clock()

			pad = 87-len(msg)
			sys.stdout.write(msg + " "*pad)			

		else:
			timer = (time.clock() - t1)						#seconds since last reading		
			#logging.debug("timer: " + str(timer) + "s")
			if timer >= hbtimeout:							#check if it's been too long since we got a reading
				logging.warning("Timeout waiting for heartbeat! ")
				#timeout = True
				break	#break out of while; this returns to start() where re-connect will be attempted		
		time.sleep(0.1)	#wait to keep the program from using too many resources continually polling the serial port

	#return	
	#if not quitFlag:		#restart only if timeout has occurred and no quit flag
	#	restart()			#attempt to re-establish connection		
	#otherwise, we exit the loop (quit condition)
		
#execution starts here

mainThread = threading.Thread(target=start)
mainThread.daemon = False 
checkKeyThread = threading.Thread(target=checkKey)
checkKeyThread.daemon = False 
setup()
checkKeyThread.start()
mainThread.start()


